package com.tong.redispool;

import org.springframework.stereotype.Component;
import redis.clients.jedis.Jedis;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author: tongCheng
 * @Package: jedisPool
 * @ClassName: MyRedisPool
 * @date: 2023-10-29 15:07
 * Redis连接池
 **/
@Component
public class MyRedisPool implements IRedisPool{

    private int maxTotal;
    private long maxWaitMillis;

    //定义一个空闲的连接集合
    private LinkedBlockingQueue<Jedis> idleObjects = null;
    //定义一个活动的连接集合
    private LinkedBlockingQueue<Jedis> activeObjects = null;
    //总的redis连接数量 原子操作类
    private AtomicInteger count = new AtomicInteger();
    @Override
    public void init(int maxTotal, long maxWaitMillis) {
        this.maxTotal = maxTotal;
        this.maxWaitMillis = maxWaitMillis;
        idleObjects = new LinkedBlockingQueue<Jedis>(maxTotal);
        activeObjects = new LinkedBlockingQueue<Jedis>(maxTotal);
    }

    @Override
    public Jedis getResource() throws Exception {
        //1.记录开始时间戳，判断超时时间
        long startTime = System.currentTimeMillis();
        //2.从空闲连接集合中获取连接，如果获取到了一个空闲的连接，将连接放到活动连接。
        Jedis redis = null;
        while (redis == null) {
            redis = idleObjects.poll();
            if (redis != null) {
                activeObjects.offer(redis);
                return redis;
            }
            //3.如果无法从空闲连接集合中拿到连接，判断连接池是否已满，如果没有，创建一个新的连接，并放入活动连接，返回。
            if (count.get() < maxTotal) {
                if (count.incrementAndGet() <= maxTotal) {
                    redis = new Jedis("127.0.0.1", 6379);
                    System.out.println("----------Redis连接池创建了一个连接");
                    activeObjects.offer(redis);
                    return redis;
                } else {
                    count.decrementAndGet();
                }
            }
            //4.如果连接池满了，等待其他线程释放连接到空闲连接集合中。如果一定时间内，可以得到一个空闲的连接，将连接放入活动连接集合中，返回
            try {
                redis = idleObjects.poll(maxWaitMillis - (System.currentTimeMillis() - startTime),
                        TimeUnit.MILLISECONDS);
                if (redis != null) {
                    activeObjects.offer(redis);
                    return redis;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
            //5.如果等待时间超过了我们最大等待超时时间，抛出等待超时异常
            if (maxWaitMillis < (System.currentTimeMillis() - startTime)) {
                throw new Exception("timeout，超过最长等待连接时间。");
            }else {
                continue;
            }
        }
        return redis;
    }

    @Override
    public void release(Jedis jedis) {
        if (activeObjects.remove(jedis)) {
            idleObjects.offer(jedis);
        }
    }
}
